require 'forwardable'
require 'observer'

module StateDesignPattern

  class StateMachine
    extend Forwardable
    include Observable

    attr_reader :context

    def initialize
      initialize_context
      transition_to_start_state
      setup_delegation
    end

    def initial_context
    end

    def start_state
      raise NotImplementedError
    end

    def current_state
      @state.class
    end

    def transition_to_state_and_send_event(state_class, name, message = {})
      transition_to_state(state_class)
      send_event(name, message)
    end

    def transition_to_state(state_class)
      @state = state_class.new(self)
    end

    def send_event(name, message = {})
      event = { name: name, source: self }.merge(message)

      changed
      notify_observers(event)

      self
    end

    private

      def initialize_context
        @context = initial_context
      end

      def transition_to_start_state
        transition_to_state(start_state)
      end

      def setup_delegation
        setup_context_delegation
        setup_action_delegation
      end

      def setup_context_delegation
        if @context
          methods = @context.members.map { |reader| [reader, "#{reader}=".to_sym] }.flatten
          self.class.def_delegators :@context, *methods
        end
      end

      def setup_action_delegation
        if @state.respond_to?(:actions)
          methods = @state.actions
          self.class.def_delegators :@state, *methods
        end
      end
  end
end
